/*
S* PORTB: LCD D0-D7
 * PORTD: PD0:SWITCH, PD1:TXD, PD2(INT0):pulse input, PD5(OC0B): beep out, PD3,PD4,PD6:LCD control R/W,E,RS
 */
#define F_CPU 10000000UL
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <stdlib.h>
#include <util/delay.h>
#include "xitoa.h"

#define NO_DEAD_TIME	1

#define SYSCLK	F_CPU
#define	BAUD	9600
#define BRR	(SYSCLK / 16 / BAUD - 1)

#define TIMER0_200US	(SYSCLK / 64 / 5000 - 1)	/* 5KHz */
#define TIMER0_4KHZ	(SYSCLK / 64 / 4000 - 1)

#define TIMER1_100HZ	(SYSCLK / 8 / 100 - 1)

#define LCD_RS	6	
#define LCD_RW	3
#define LCD_E	4

#define LCD_CMD_CLR	1
#define LCD_CMD_HOME	2
#define LCD_CMD_MODESET	4
#define LCD_CMD_DISPLAY	8
#define LCD_CMD_CURSOR	0x10
#define LCD_CMD_FUNCTION	0x20
#define LCD_CMD_CGRAM	0x40
#define LCD_CMD_DDRAM	0x80

#define LCD_RS_DATA	_BV(LCD_RS)

void update_display(void);

void lcd_init(void);
void lcd_write(char, unsigned char);
void lcd_clr(void);
void lcd_nl(void);
void lcd_putc(unsigned char c);
void uart_init(void);
void uart_putc(unsigned char c);

#define EVENT_TIMEOUT	1
#define EVENT_SW_ON	2
#define EVENT_SW_LONG	4
volatile unsigned char event;
unsigned char tick_10ms;
unsigned char switch_state;

volatile unsigned char cps, last_cps;
volatile unsigned int last_cp2s;
unsigned char max_cps;
volatile unsigned int cpm, max_cpm;

unsigned char display_mode;

#define LOG_STEP	2
#define LOG_SIZE	(60 / LOG_STEP)
unsigned char counts_log[LOG_SIZE];
unsigned char log_p;
unsigned char counts_for_log;
unsigned char log_step;

unsigned long counts_total;
unsigned long seconds;

/* 
 *beep driver
 */
unsigned char beep_on;

static inline void beep_stop()
{
	TCCR0B = 0;
	beep_on = 0;
	DDRD &= ~0x20;
}

static inline void beep_start()
{
	TCCR0B = 3;	/* CK/8 */
	TCNT0 = 0;
	beep_on = 10;	/* 100msec */
	DDRD |= 0x20;
}

static inline void beep_update()
{
	if (beep_on > 0) {
		beep_on--;
		if (beep_on == 0) {
			TCCR0B = 0;	/* stop beep timer */
			DDRD &= ~0x20;
		}
	}
}

/* check if it passed enough time from the beep started */
static inline char beep_check_time(void)
{
	if (beep_on == 10 && TCNT0 < TIMER0_200US)
		return 0;	/* < 200us */
	else
		return 1;
}
/*
 * switch
 */

static inline char switch_is_on(void)
{
	return ((PIND & _BV(PIND0)) == 0) ? 1 : 0;
}


void clear_logs(void)
{
	cli();

	cps = 0;
	last_cps = 0;
	last_cp2s = 0;
	max_cps = 0;
	cpm = 0;
	max_cpm = 0;
	for (log_p = 0; log_p < LOG_SIZE; log_p++)
		counts_log[log_p] = 0;
	log_p = 0;

	counts_total = 0;
	seconds = 0;

	counts_for_log = 0;
	log_step = 0;

	sei();
}


void init_ports(void)
{
	cli();

	/* PORTB: output mode */
	PORTB = 0x00;
	DDRB = 0xff;

	/* PORTD:  PD3,PD4,PD6:out PD0:pull-up */
	PORTD = 0x01;
	DDRD = 0x78;

	/* INT0: falling edge */
	MCUCR = 0x02;		

	GIMSK = 0x40;	/* enable INT0 */

	sei();
}

void init_timers()
{
	cli();

	beep_on = 0;
	tick_10ms = 0;
	switch_state = 0;

	TIMSK = 0;
	/* 8 bit timer for beep */
	TCCR0A = 0x12;	/* toggle OC0B, CTC mode */
	TCCR0B = 0;	/* stop */
	TCNT0 = 0;
	OCR0A = TIMER0_4KHZ;
		

	/* 16bit timer for counting 10m seconds */
	TCCR1A = 0x00;
	TCCR1B = 0x0a;	/* CTC mode(WGM12), CK/8 */
	TCNT1 = 0;
	OCR1A = TIMER1_100HZ;

	TIMSK = 0x40;	/* enable timer counter 1 compare A match interrupt */

	sei();
}

/* 10ms timer */
ISR(SIG_TIMER1_COMPA)
{
	tick_10ms++;
	if (tick_10ms == 100) {
		tick_10ms = 0;

		/* save and reset current cps */
		last_cp2s = last_cps + cps;
		last_cps = cps;
		cps = 0;

		event |= EVENT_TIMEOUT;
	}

	/* update beep state */
	beep_update();

	if (switch_is_on()) {
		if (switch_state == 0)
			event |= EVENT_SW_ON;
		else if (switch_state == 200)	/* 2sec */
			event |= EVENT_SW_LONG;
		switch_state++;
	} else {
		/* released */
		switch_state = 0;
	}
}

ISR(SIG_INTERRUPT0)
{
#ifndef NO_DEAD_TIME
	if (!beep_check_time()) {
		/* called too quickly */
		return;	/* ignore */
	}
#endif
	if (cps < 255)
		cps++;
	
	beep_start();
}


int main()
{
	uart_init();

	event = 0;
	display_mode = 0;

	clear_logs();
	init_ports();
	init_timers();

	lcd_init();
	lcd_clr();

	xdev_out(lcd_putc);

	for (;;) {
		update_display();

		/* wait for the next period */
		while (event == 0)
			;
		
		cli();
		if (event & EVENT_TIMEOUT) {
			/* update counting registers */
			event &= ~EVENT_TIMEOUT;

			sei();

			seconds++;

			if (counts_for_log + last_cps < counts_for_log)
				counts_for_log = 0xff;
			else
				counts_for_log = counts_for_log + last_cps;

			log_step++;
			if (log_step == LOG_STEP) {
				log_step = 0;
				cpm = cpm - counts_log[log_p] + counts_for_log;
				counts_log[log_p++] = counts_for_log;
				if (log_p == LOG_SIZE)
					log_p = 0;
				counts_for_log = 0;
			}

			if (counts_total + last_cps < counts_total)
				counts_total = 0xffffffff;
			else
				counts_total += last_cps;

			if (last_cps > max_cps)
				max_cps = last_cps;
			if (cpm > max_cpm)
				max_cpm = cpm;

			xdev_out(uart_putc);
			xitoa(seconds, 10, 0);
			xputc(' ');
			xitoa(last_cps, 10, 0);
			xputc(' ');
			xitoa(cpm, 10, 0);
			xputc(' ');
			xitoa(counts_total, 10, 0);
			xputc('\r');
			xputc('\n');

		} else if (event & EVENT_SW_ON) {
			event &= ~EVENT_SW_ON;

			display_mode++;
			if (display_mode > 3)
				display_mode = 0;

		} else if (event & EVENT_SW_LONG) {
			event &= ~EVENT_SW_LONG;

			init_timers();
			clear_logs();
			display_mode = 0;
		}
		sei();
	}

	return 0;
}


void
update_display()
{
	lcd_clr();
	xdev_out(lcd_putc);

	xitoa(last_cps, 10, 3);
	xputs(PSTR(" cps"));
	xitoa(cpm, 10, 5);
	xputs(PSTR(" cpm"));
	lcd_nl();

	switch (display_mode) {
	case 0:
		xitoa(counts_total, 10, 5);
		xputs(PSTR(" /"));
		xitoa(seconds, 10, 5);
		xputs(PSTR(" sec"));
		break;
	case 1:
		xitoa(max_cps, 10, 3);
		xitoa(max_cpm, 10, 9);
		xputs(PSTR(" max"));
		break;
#if 0
	case 2: /* Sv=0.8Gy */
		xitoa((last_cp2s * 69) >> 5, 10, 3);
		xputs(PSTR(" uSv"));
		xitoa((cpm * 123) >> 1, 10, 5);
		xputs(PSTR(" nSv"));
		break;
#else
	case 2:	/* Sv=Gy */
		xitoa((last_cp2s * 43) >> 4, 10, 3);
		xputs(PSTR(" uSv"));
		xitoa(cpm * 77, 10, 5);
		xputs(PSTR(" nSv"));
		break;
#endif
	default:
	case 3:
		xitoa((last_cp2s * 85) >> 8, 10, 3);
		xputs(PSTR(" mR/h"));
		xitoa((cpm * 35) >> 2, 10, 4);
		xputs(PSTR("uR/h"));
		break;
	}
}


/*
 * LCD driver
 */
void
lcd_write(char data, unsigned char c)
{
	
	PORTB = c;
	PORTD &= ~_BV(LCD_RS);
	PORTD |= data;
	PORTD |= _BV(LCD_E);
	asm("nop");
	asm("nop");
	asm("nop");
	PORTD &= ~_BV(LCD_E);
}

void
lcd_clr(void)
{
	lcd_write(0, LCD_CMD_CLR);
	_delay_ms(2);
}


void
lcd_nl(void)
{
	lcd_write(0, LCD_CMD_DDRAM | 0x40);
	_delay_us(40);
}

void
lcd_putc(unsigned char c)
{
	lcd_write(LCD_RS_DATA, c);
	_delay_us(40);
}

void
lcd_init(void)
{
	_delay_ms(10);
	lcd_write(0, 0x30);
	_delay_ms(5);
	lcd_write(0, 0x30);
	_delay_us(100);
	lcd_write(0, 0x30);
	_delay_us(40);

	lcd_write(0, LCD_CMD_FUNCTION | 0x18);	/* 8bit, 2lines */
	_delay_us(40);
	lcd_write(0, LCD_CMD_DISPLAY);	/* display off, cursor off, blink off */
	_delay_us(40);
	lcd_write(0, LCD_CMD_MODESET | 2);	/* inclement */
	_delay_us(40);
	lcd_write(0, LCD_CMD_DISPLAY | 0x4);	/* display on */
	_delay_us(40);
}

/*
 & UART driver
 */
void uart_init(void)
{
	UBRRH = (unsigned char)(BRR >> 8);
	UBRRL = (unsigned char)BRR;
	UCSRB = _BV(TXEN);
	UCSRC = _BV(UCSZ0) | _BV(UCSZ1);
}

void uart_putc(unsigned char c)
{
	loop_until_bit_is_set(UCSRA, UDRE);
	UDR = c;
}

